/**
 * @brief Ashikhmin Tone Mapping Operator
 * 
 * Ashikhmin tone reproduction for displaying high contrast
 * scenes. 
 * Michael Ashikhmin 2002
 *
 * This file is a part of PFSTMO package.
 * ---------------------------------------------------------------------- 
 * Copyright (C) 2003,2004 Grzegorz Krawczyk
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ---------------------------------------------------------------------- 
 * 
 * @author Akiko Yoshida, <yoshida@mpi-sb.mpg.de>
 * @author Grzegorz Krawczyk, <krawczyk@mpi-sb.mpg.de>
 *
 * $Id: pfstmo_ashikhmin02.cpp,v 1.5 2004/12/15 10:11:03 krawczyk Exp $
 */

#include <config.h>

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <math.h>

#include <pfs.h>

#include <tmo_ashikhmin02.h>

#define PROG_NAME "pfstmo_ashikhmin02"

using namespace std;

class QuietException 
{
};

void calculateLuminance( pfs::Array2D* Y, float& avLum, float& maxLum, float& minLum);


void printHelp()
{
  fprintf( stderr, PROG_NAME ": \n"
    "\t [--simple] \n"
    "\t [--local-contrast <val>] [--eq <num>] \n"
    "\t [--verbose] [--help] \n"
    "See man page for more information. \n" );
}

void pfstmo_ashikhmin02( int argc, char* argv[] )
{
  pfs::DOMIO pfsio;

  //--- default tone mapping parameters;
  bool simple_flag = false;
  float lc_value = 0.5;
  int eq = 2;

  //--- process command line args
  bool verbose = false;

  static struct option cmdLineOptions[] = {
    { "help", no_argument, NULL, 'h' },
    { "verbose", no_argument, NULL, 'v' },
    { "simple", no_argument, NULL, 's' },
    { "local-contrast", required_argument, NULL, 'l' },
    { "equation", required_argument, NULL, 'e' },
    { NULL, 0, NULL, 0 }
  };

  int optionIndex = 0;
  while( 1 ) {
    int c = getopt_long (argc, argv, "hvsl:e:", cmdLineOptions, &optionIndex);
    if( c == -1 ) break;
    switch( c ) {
    case 'h':
      printHelp();
      throw QuietException();
    case 'v':
      verbose = true;
      break;
    case 's': // TM function only
      simple_flag = true;
      break;
    case 'l': // local contrast threshold
      lc_value = (float)strtod(optarg, NULL);
      break;
    case 'e': // equation on the paper
      eq = (int) strtod(optarg, NULL);
      if( eq!=2 && eq!=4 )
        throw pfs::Exception("Incorrect equation number, accepted value is 2 or 4");
      break;
    case '?':
      throw QuietException();
    case ':':
      throw QuietException();
    }
  }
  
  if( simple_flag )
  {
    VERBOSE_STR << "applying the tone mapping function only" << endl;
  }
  else
  {
    VERBOSE_STR << "local contrast threshold: " << lc_value << endl;
    VERBOSE_STR << "equation used: " << eq << ". " << endl;
  }
   
  while( true ) {
    pfs::Frame *frame = pfsio.readFrame( stdin );
    if( frame == NULL ) break; // No more frames

    pfs::Channel *X, *Y, *Z;
    frame->getXYZChannels( X, Y, Z );
    //---

    if( Y == NULL )
      throw pfs::Exception( "Missing X, Y, Z channels in the PFS stream" );
        
    float maxLum,avLum, minLum;
    calculateLuminance( Y, avLum, maxLum, minLum);
    /*
     VERBOSE_STR << "maximum luminance: " << maxLum << endl;
     VERBOSE_STR << "minumum luminance: " << minLum << endl;
     VERBOSE_STR << "average luminance: " << avLum << endl;
    */

    int w = Y->getCols();
    int h = Y->getRows();

    pfs::Array2D* L = new pfs::Array2DImpl(w,h);
    tmo_ashikhmin02(Y, L, maxLum, minLum, avLum, simple_flag, lc_value, eq);
		
    for( int x=0 ; x<w ; x++ )
      for( int y=0 ; y<h ; y++ )
      {
        float scale = (*L)(x,y) / (*Y)(x,y);
        (*Y)(x,y) *= scale;
        (*X)(x,y) *= scale;
        (*Z)(x,y) *= scale;
      }

    delete L;

    //---
    pfsio.writeFrame( frame, stdout );
    pfsio.freeFrame( frame );        
  }
}

void calculateLuminance( pfs::Array2D* Y, float& avLum, float& maxLum, float& minLum)
{
  avLum = 0.0f;
  maxLum = 0.0f;
  minLum = 0.0f;

  int size = Y->getCols() * Y->getRows();

  for( int i=0 ; i<size; i++ )
  {
    avLum += log( (*Y)(i) + 1e-4 );
    maxLum = ( (*Y)(i) > maxLum ) ? (*Y)(i) : maxLum ;
    minLum = ( (*Y)(i) < minLum ) ? (*Y)(i) : minLum ;
  }
  avLum =exp( avLum/ size);
}


int main( int argc, char* argv[] )
{
  try {
    pfstmo_ashikhmin02( argc, argv );
  }
  catch( pfs::Exception ex ) {
    fprintf( stderr, PROG_NAME " error: %s\n", ex.getMessage() );
    return EXIT_FAILURE;
  }        
  catch( QuietException  ex ) {
    return EXIT_FAILURE;
  }        
  return EXIT_SUCCESS;
}
